from dataclasses import dataclass, field
from typing import ClassVar, cast

from pytest import raises

from apischema.fields import FIELDS_SET_ATTR
from apischema.validation.mock import NonTrivialDependency, ValidatorMock


@dataclass
class Data:
    a: int
    b: str = "1"
    c: ClassVar[int] = 42
    d = 0

    @property
    def property(self) -> int:
        return int(self.b) + self.a

    def method(self, arg: int) -> int:
        return self.a + arg

    @classmethod
    def classmethod(cls, arg: int):
        return cls.c + arg


def test_mock():
    mock = cast(Data, ValidatorMock(Data, {"a": 0}, {"b": field(default="1")}))
    assert mock.a == 0
    assert mock.b == "1"
    assert mock.c == 42
    assert mock.__class__ == Data
    assert mock.__dict__ == {"a":             0, "b": "1",
                             FIELDS_SET_ATTR: {"a"}}
    assert mock.property == 1
    assert mock.method(1) == 1
    assert mock.classmethod(0) == 42
    assert type(mock) == ValidatorMock
    # limitation of mocking, class variables must be typed
    # in order to be detected
    with raises(NonTrivialDependency):
        _ = mock.d
    with raises(NonTrivialDependency):
        _ = mock.e



from dataclasses import dataclass, field
from typing import NewType

from pytest import raises

from apischema import (ValidationError, from_data, output_converter,
                       to_data, validator)

Password = NewType("Password", str)


@output_converter
def hide_password(_: Password) -> str:
    return "******"


@dataclass
class ChangePasswordForm:
    password: Password = field()
    confirmation: Password

    @validator
    def password_match_confirmation(self):
        if self.password != self.confirmation:
            yield "password and its confirmation don't match"


def test_change_password_form():
    data = {
        "password":     "5tr0ngP455w0rd!",
        "confirmation": "5tr0ngP455w0rd!"
    }
    form = from_data(ChangePasswordForm, data)
    assert form == ChangePasswordForm(Password("5tr0ngP455w0rd!"),
                                      Password("5tr0ngP455w0rd!"))
    assert to_data(form) == {
        "password":     "******",
        "confirmation": "******",
    }


def test_bad_confirmation():
    data = {
        "password":     "password",
        "confirmation": "1234",
    }
    with raises(ValidationError) as err:
        from_data(ChangePasswordForm, data)
    assert err.value == ValidationError([
        "password and its confirmation don't match",
    ])


def test_missing_confirmation():
    data = {
        "password": "password",
    }
    with raises(ValidationError) as err:
        from_data(ChangePasswordForm, data)
    assert err.value == ValidationError(children={
        "confirmation": ValidationError(["missing field"])
    })



from dataclasses import (Field, MISSING, _FIELD, _FIELDS,  # type: ignore
                         _FIELD_INITVAR, fields as fields_,
                         is_dataclass)
from functools import wraps
from inspect import signature
from typing import (AbstractSet, Any, Type,
                    TypeVar, Union, cast)

from apischema.utils import PREFIX


class NoDefault(Exception):
    pass


def get_default(field: Field) -> Any:
    if field.default_factory is not MISSING:  # type: ignore
        return field.default_factory()  # type: ignore
    if field.default is not MISSING:
        return field.default
    raise NoDefault()


FIELDS_SET_CACHE_ATTR = f"{PREFIX}fields_set_cache"

FIELDS_SET_ATTR = f"{PREFIX}fields_set"

Cls = TypeVar("Cls", bound=Type)


def check_dataclass(obj):
    if not is_dataclass(obj):
        raise ValueError("not a dataclass")


def with_fields_set(cls: Cls) -> Cls:
    check_dataclass(cls)
    init_fields = set()
    post_init_fields = set()
    for field in getattr(cls, _FIELDS).values():
        assert isinstance(field, Field)
        if field._field_type == _FIELD_INITVAR:  # type: ignore
            init_fields.add(field.name)
        if field._field_type == _FIELD and not field.init:  # type: ignore
            post_init_fields.add(field.name)
    params = list(signature(cls.__init__).parameters)
    old_new = cls.__new__
    old_init = cls.__init__
    old_setattr = cls.__setattr__

    @wraps(old_new)
    def new_new(*args, **kwargs):
        if old_new is object.__new__:
            obj = object.__new__(args[0])
        else:
            obj = old_new(*args, **kwargs)
        obj.__dict__[FIELDS_SET_ATTR] = set()
        return obj

    @wraps(old_init)
    def new_init(*args, **kwargs):
        args[0].__dict__[FIELDS_SET_ATTR] = set()
        old_init(*args, **kwargs)
        arg_fields = {*params[1:len(args)], *kwargs} - init_fields
        args[0].__dict__[FIELDS_SET_ATTR] = arg_fields | post_init_fields

    @wraps(old_setattr)
    def new_setattr(self, attr, value):
        self.__dict__[FIELDS_SET_ATTR].add(attr)
        old_setattr(self, attr, value)

    cls.__new__ = new_new
    cls.__init__ = new_init
    cls.__setattr__ = new_setattr
    return cls


T = TypeVar("T")


def mark_set_fields(obj: T, *fields: str, overwrite=False) -> T:
    check_dataclass(obj)
    all_fields = {f.name for f in fields_(obj)}
    if any(f not in all_fields for f in fields):
        raise ValueError(f"{set(fields) - all_fields} are not fields")
    if overwrite:
        obj.__dict__[FIELDS_SET_ATTR] = set(fields)
    else:
        try:
            obj.__dict__[FIELDS_SET_ATTR].update(fields)
        except KeyError:
            fs = get_fields_set(obj)
            if any(f not in fs for f in fields):
                obj.__dict__[FIELDS_SET_ATTR] = {*fs, *fields}
    return obj


def unmark_set_fields(obj: T, *fields: str) -> T:
    check_dataclass(obj)
    if FIELDS_SET_ATTR in obj.__dict__:
        obj.__dict__[FIELDS_SET_ATTR].difference_update(fields)
    else:
        obj.__dict__[FIELDS_SET_ATTR] = set(get_fields_set(obj)).difference(fields)
    return obj


def get_fields_set(obj: Any) -> AbstractSet[str]:
    check_dataclass(obj)
    try:
        return getattr(obj, FIELDS_SET_ATTR)
    except AttributeError:
        return {f.name for f in fields_(obj)}


class FieldGetter:
    def __init__(self, obj):
        check_dataclass(obj)
        self.obj = obj
        self.fields = {f.name: f for f in fields_(obj)}

    def __getattribute__(self, attr) -> Field:
        try:
            return super().__getattribute__("fields")[attr]
        except KeyError:
            cls = type(super().__getattribute__("obj")).__name__
            raise AttributeError(f"Class {cls} has no field 'attr'")


def get_fields(obj: Union[Type[T], T]) -> T:
    return cast(T, FieldGetter(obj))



from dataclasses import dataclass
from typing import Any, Dict, Optional, Sequence, Union

from apischema.utils import NO_DEFAULT, PREFIX

ANNOTATIONS_METADATA = f"{PREFIX}description"


@dataclass
class Annotations:
    title: Optional[str] = None
    description: Optional[str] = None
    default: Optional[Any] = NO_DEFAULT
    examples: Optional[Sequence[Any]] = None
    format: Optional[str] = None
    items: Optional[Union["Annotations", Sequence["Annotations"]]] = None
    additional_properties: Optional["Annotations"] = None


_annotations: Dict[Any, Annotations] = {}

get_annotations = _annotations.get



from dataclasses import is_dataclass, replace
from enum import Enum
from itertools import chain
from typing import (Any, Callable, Dict, Iterable, List, Mapping, Optional, Pattern,
                    Sequence,
                    Set, Tuple, Type, Union)

from apischema.conversion import (Converter, InputVisitorMixin,
                                  OutputVisitorMixin)
from apischema.data import to_data
from apischema.dataclasses import (Field, FieldCache, FieldKind, get_input_fields,
                                   get_output_fields)
from apischema.fields import (NoDefault, get_default, get_fields_set, mark_set_fields)
from apischema.ignore import Ignored
from apischema.json_schema.types import JSONSchema, JSONType
from apischema.schema import Constraint, Schema
from apischema.schema.annotations import get_annotations
from apischema.schema.constraints import (ArrayConstraint, NumberConstraint,
                                          ObjectConstraint, StringConstraint,
                                          get_constraint)
from apischema.utils import NO_DEFAULT, as_dict
from apischema.visitor import Visitor

constraint_by_type = {
    int:   NumberConstraint,
    float: NumberConstraint,
    str:   StringConstraint,
}


class ReadWriteOnly(Constraint):
    pass


def check_constraint(schema: Schema, expected: Type[Constraint]):
    if schema.constraint is not None:
        if not isinstance(schema.constraint, expected):
            if not isinstance(schema.constraint, ReadWriteOnly):
                raise TypeError(f"Bad constraint: expected {expected.__name__}"
                                f" found {type(schema.constraint).__name__}")


def _remove_none(obj: Optional[Any]) -> Iterable[Tuple[str, Any]]:
    if obj is None:
        return iter(())
    return ((k, v) for k, v in as_dict(obj).items()
            if v is not None and v is not NO_DEFAULT
            and k not in ("items", "additional_properties"))


def _to_dict(schema: Schema) -> Mapping[str, Any]:
    return dict(chain(_remove_none(schema.annotations),
                      _remove_none(schema.constraint)))


def _merge_constraints(c1: Optional[Constraint],
                       c2: Optional[Constraint]) -> Optional[Constraint]:
    if c1 is None:
        return c2
    if c2 is None:
        return c1
    if isinstance(c1, ReadWriteOnly):
        return type(c2)(**{**as_dict(c1), **as_dict(c2)})
    elif Constraint != type(c1) != type(c2) != Constraint:
        raise TypeError(f"Incompatibles constraints {type(c1).__name__}"
                        f" and {type(c2).__name__}")
    return c1


def _override(schema: Schema, cls: Type) -> Schema:
    return Schema(schema.annotations or get_annotations(cls),
                  _merge_constraints(schema.constraint, get_constraint(cls)))


def _field_schema(field: Field) -> Schema:
    annotations = field.annotations
    if annotations is not None and annotations.default is ...:
        try:
            default = to_data(get_default(field.base_field))
        except NoDefault:
            raise TypeError("Invalid ... without field default")
        annotations = replace(annotations, default=default)
    return Schema(annotations, field.constraint)


def _extract_properties_schema(json_schema: JSONSchema) -> JSONSchema:
    if json_schema.pattern_properties is not None:
        if (
                len(json_schema.pattern_properties) >= 1
                or json_schema.additional_properties
        ):  # don't try to merge the schemas and return
            return JSONSchema()
        return next(iter(json_schema.pattern_properties.values()))
    if json_schema.additional_properties:
        if isinstance(json_schema.additional_properties, JSONSchema):
            return json_schema.additional_properties
        else:
            return JSONSchema()
    raise TypeError("properties field must have an 'object' schema")


class SchemaBuilder(Visitor[Schema, JSONSchema]):
    def __init__(self, ref_factory: Optional[Callable[[Type], str]]):
        super().__init__()
        self.ref_factory = ref_factory or self._missing_ref_factory
        self.schemas: Set[Type] = set()

    @staticmethod
    def _missing_ref_factory(cls: Type) -> str:
        raise TypeError("reference factory is needed to handle"
                        " recursive types")

    def primitive(self, cls: Type, schema: Schema) -> JSONSchema:
        if cls in constraint_by_type:
            check_constraint(schema, constraint_by_type[cls])
        return JSONSchema(type=JSONType.from_type(cls), **_to_dict(schema))

    def union(self, alternatives: Sequence[Type],
              schema: Schema) -> JSONSchema:
        any_of = []
        for cls in alternatives:
            try:
                any_of.append(self.visit(cls, schema))
            except Ignored:
                pass
        if len(any_of) == 1:
            return any_of[0]
        else:
            return JSONSchema(any_of=any_of)

    def iterable(self, cls: Type[Iterable], value_type: Type,
                 schema: Schema) -> JSONSchema:
        check_constraint(schema, ArrayConstraint)
        if isinstance(schema.items_, Sequence):
            raise TypeError("Iterable items schema cannot be a list of schema")
        return JSONSchema(type=JSONType.ARRAY,
                          items=self.visit(value_type, schema.items_),
                          **_to_dict(schema))

    def mapping(self, cls: Type[Mapping], key_type: Type, value_type: Type,
                schema: Schema) -> JSONSchema:
        check_constraint(schema, ObjectConstraint)
        key = self.visit(key_type, Schema())
        pattern = key.pattern
        value = self.visit(value_type, schema.additional_properties)
        properties: Dict[str, Any] = {}
        if key.pattern is not None:
            properties["pattern_properties"] = {key.pattern: value}
        else:
            properties["additional_properties"] = value
        return JSONSchema(type=JSONType.OBJECT, **properties,  # type: ignore
                          **_to_dict(schema))

    def typed_dict(self, cls: Type, keys: Mapping[str, Type], total: bool,
                   schema: Schema) -> JSONSchema:
        if cls in self.schemas:
            return JSONSchema(ref=self.ref_factory(cls))
        check_constraint(schema, ObjectConstraint)
        if schema.additional_properties:
            raise TypeError("additional properties are not handled"
                            " for TypedDict")
        sorted_keys = sorted(keys)
        return JSONSchema(type=JSONType.OBJECT,
                          properties={key: self.visit(keys[key], Schema())
                                      for key in sorted_keys},
                          required=sorted_keys if total else [],
                          additional_properties=True, **_to_dict(schema))

    def tuple(self, types: Sequence[Type], schema: Schema) -> JSONSchema:
        check_constraint(schema, ArrayConstraint)
        if schema.constraint is not None:
            assert isinstance(schema.constraint, ArrayConstraint)
            if (schema.constraint.max_items is not None
                    or schema.constraint.min_items is not None):
                raise TypeError("Tuple cannot have min_items/max_items"
                                " constraint")
        if isinstance(schema.items_, Sequence):
            if len(schema.items_) != len(types):
                raise TypeError("Tuple items schema list must have the same"
                                " length than tuple")
        items = []
        for i, cls in enumerate(types):
            if isinstance(schema.items_, Sequence):
                elt_schema = schema.items_[i]
            else:
                elt_schema = schema.items_
            items.append(self.visit(cls, elt_schema))
        return JSONSchema(type=JSONType.ARRAY, items=items,
                          min_items=len(types), max_items=len(types),
                          **_to_dict(schema))

    def literal(self, values: Sequence[Any], schema: Schema) -> JSONSchema:
        if not values:
            raise TypeError("empty Literal")
        if schema.constraint is not None:
            raise TypeError("Literal cannot have constraint")
        types = sorted(set(
            JSONType.from_type(type(v.value if isinstance(v, Enum) else v))
            for v in values
        ))
        type_ = types[0] if len(types) == 1 else types
        if len(values) == 1:
            return JSONSchema(type=type_, const=values[0],  # type: ignore
                              **_to_dict(schema))
        else:
            return JSONSchema(type=type_, enum=values,  # type: ignore
                              **_to_dict(schema))

    def _dataclass_fields(self, cls: Type) -> FieldCache:
        raise NotImplementedError()

    def _field_visit(self, field: Field) -> JSONSchema:
        raise NotImplementedError()

    def dataclass(self, cls: Type, schema: Schema) -> JSONSchema:
        assert is_dataclass(cls)
        if cls in self.schemas:
            return JSONSchema(ref=self.ref_factory(cls))
        self.schemas.add(cls)
        check_constraint(schema, ObjectConstraint)
        schema_ = _override(schema, cls)
        properties = {}
        required: List[str] = []
        fields, pattern_fields, additional_field = self._dataclass_fields(cls)
        for field in fields:
            properties[field.alias] = self._field_visit(field)
            if not field.default:
                required.append(field.alias)
        pattern_properties: Dict[Pattern, JSONSchema] = {}
        for pattern, field in pattern_fields:
            field_schema = self._field_visit(field)
            pattern_properties[pattern] = _extract_properties_schema(field_schema)
        additional_properties: Optional[Union[bool, "JSONSchema"]] = False
        if additional_field:
            field_schema = self._field_visit(additional_field)
            additional_properties = _extract_properties_schema(field_schema)
        kwargs: Dict[str, Any] = {}
        if properties:
            kwargs["properties"] = properties
        if required:
            kwargs["required"] = required
        # get_fields_set(additional_properties) means that additional_properties
        # is not {} and not True
        if additional_properties is False or get_fields_set(additional_properties):
            kwargs["additional_properties"] = additional_properties
        if pattern_properties:
            kwargs["pattern_properties"] = pattern_properties
        return JSONSchema(type=JSONType.OBJECT,  # type: ignore
                          **kwargs, **_to_dict(schema_))

    def enum(self, cls: Type[Enum], schema: Schema) -> JSONSchema:
        if schema.constraint is not None:
            raise TypeError("enum cannot have constraint")
        if len(cls) == 0:
            raise TypeError("Empty enum")
        return self.literal(list(cls), _override(schema, cls))

    def new_type(self, cls: Type, super_type: Type, schema: Schema
                 ) -> JSONSchema:
        return self.visit(super_type, _override(schema, cls))

    def any(self, schema: Schema) -> JSONSchema:
        if schema.constraint is not None:
            raise TypeError("Any type cannot have constraint")
        return JSONSchema()

    def annotated(self, cls: Type, annotations: Sequence[Any],
                  schema: Schema) -> JSONSchema:
        schema_ = schema
        if Ignored in annotations:
            raise Ignored
        for annotation in annotations:
            if isinstance(annotation, Schema):
                schema_ = Schema(
                    schema.annotations or annotation.annotations,
                    _merge_constraints(schema.constraint,
                                       annotation.constraint)
                )
                break
        return self.visit(cls, schema_)


class InputSchemaBuilder(InputVisitorMixin[Schema, JSONSchema], SchemaBuilder):
    def __init__(self, ref_factory: Optional[Callable[[Type], str]],
                 additional_properties: bool):
        SchemaBuilder.__init__(self, ref_factory)
        self.additional_properties = additional_properties

    def _custom(self, cls: Type, custom: Dict[Type, Converter],
                schema: Schema) -> JSONSchema:
        return self.union(list(custom), _override(schema, cls))

    _dataclass_fields = staticmethod(get_input_fields)  # type: ignore

    def _field_visit(self, field: Field) -> JSONSchema:
        schema = _field_schema(field)
        if field.kind == FieldKind.INIT:
            if schema.constraint is None:
                schema.constraint = ReadWriteOnly(write_only=True)
            else:
                if schema.constraint.read_only:
                    raise TypeError("InitVar cannot be read-only")
                schema.constraint = replace(schema.constraint, write_only=True)
        return self.visit(field.input_type, schema)

    def dataclass(self, cls: Type, schema: Schema) -> JSONSchema:
        json_schema = super().dataclass(cls, schema)
        if json_schema.additional_properties is False:
            # TODO document that `dataclasses.replace` don't keep fields set
            return mark_set_fields(replace(
                json_schema, additional_properties=self.additional_properties
            ), *get_fields_set(json_schema), overwrite=True)
        return json_schema


def build_input_schema(cls: Type, *, ref_factory: Callable[[Type], str] = None,
                       additional_properties: bool = False,
                       schema: Schema = None) -> JSONSchema:
    builder = InputSchemaBuilder(ref_factory, additional_properties)
    return builder.visit(cls, schema or Schema())


class OutputSchemaBuilder(OutputVisitorMixin[Schema, JSONSchema],
                          SchemaBuilder):
    def __init__(self, conversions: Mapping[Type, Type],
                 ref_factory: Optional[Callable[[Type], str]]):
        SchemaBuilder.__init__(self, ref_factory)
        OutputVisitorMixin.__init__(self, conversions)

    def _custom(self, cls: Type, custom: Tuple[Type, Converter],
                schema: Schema) -> JSONSchema:
        return self.visit(custom[0], _override(schema, cls))

    _dataclass_fields = staticmethod(get_output_fields)  # type: ignore

    def _field_visit(self, field: Field) -> JSONSchema:
        schema = _field_schema(field)
        if field.kind == FieldKind.NO_INIT:
            if schema.constraint is None:
                schema.constraint = ReadWriteOnly(read_only=True)
            else:
                if schema.constraint.write_only:
                    raise TypeError("not init field cannot be write-only")
                schema.constraint = replace(schema.constraint, read_only=True)
        return self.visit(field.output_type, schema)


def build_output_schema(cls: Type, conversions: Mapping[Type, Type] = None, *,
                        ref_factory: Callable[[Type], str] = None,
                        schema: Schema = None) -> JSONSchema:
    builder = OutputSchemaBuilder(conversions or {}, ref_factory)
    return builder.visit(cls, schema or Schema())



from dataclasses import InitVar, dataclass, field

from apischema import (build_input_schema, build_output_schema, from_data,
                       schema, to_data)


@dataclass
class Data:
    init_field: InitVar[int]
    write_only_field: int = field(metadata=schema(write_only=True))
    not_init_field: int = field(init=False)
    read_only_field: int = field(default=42, metadata=schema(read_only=True))

    def __post_init__(self, init_field: int):
        self.not_init_field = init_field


def test_data():
    assert to_data(build_input_schema(Data)) == {
        "type":                 "object",
        "required":             ["init_field", "write_only_field"],
        "additionalProperties": False,
        "properties":           {
            "init_field":       {
                "type":      "integer",
                "writeOnly": True,
            },
            "write_only_field": {
                "type":      "integer",
                "writeOnly": True,
            },
            "read_only_field":  {
                "type":     "integer",
                "readOnly": True,
            }
        },
    }
    assert to_data(build_output_schema(Data)) == {
        "type":                 "object",
        "required":             ["write_only_field", "not_init_field"],
        "additionalProperties": False,
        "properties":           {
            "write_only_field": {
                "type":      "integer",
                "writeOnly": True,
            },
            "not_init_field":   {
                "type":     "integer",
                "readOnly": True,
            },
            "read_only_field":  {
                "type":     "integer",
                "readOnly": True,
            }
        },
    }
    data = from_data(Data, {"init_field": 0, "write_only_field": 1})
    assert data.write_only_field == 1
    assert data.not_init_field == 0
    assert data.read_only_field == 42
    # Support of readOnly/writeOnly is provisional because apischema works
    # with dual schema build_input_field/build_output_field (because of
    # conversions).
    # I could filter writeOnly properties at serialization (and readOnly at
    # deserialization), but this would make the code way more complex and I'm
    # not sure of the real interest of this "feature"
    assert to_data(data) != {
        "not_init_field":  0,
        "read_only_field": 42
    }
    assert to_data(data) == {
        "write_only_field": 1,
        "not_init_field":   0,
        "read_only_field":  42
    }



from typing import Generic, List, Tuple, TypeVar

from pytest import raises

from apischema import (ValidationError, build_input_schema, build_output_schema,
                       from_data,
                       input_converter,
                       output_converter, to_data)
from apischema.visitor import Unsupported

T = TypeVar("T")


class Wrapper(Generic[T]):
    def __init__(self, wrapped: T):
        self.wrapped = wrapped

    @output_converter
    def _wrapped(self) -> T:
        return self.wrapped


U = TypeVar("U")


@input_converter
def to_wrapper(a: U) -> Wrapper[U]:
    return Wrapper(a)


def test_wrapper():
    assert from_data(Wrapper[List[int]], [0, 1]).wrapped == [0, 1]
    assert to_data(Wrapper("wrapped")) == "wrapped"
    assert to_data(build_input_schema(Wrapper[int])) == {
        "type": "integer"
    } == to_data(build_output_schema(Wrapper[int]))

    with raises(ValidationError):
        from_data(Wrapper[int], "wrapped")


##########################################


class Pair(Generic[T, U]):
    def __init__(self, a: T, b: U):
        self.a = a
        self.b = b


@input_converter
def pair(value: Tuple[int, T]) -> Pair[int, T]:
    a, b = value
    return Pair(a, b)


def test_pair():
    with raises(Unsupported):
        from_data(Pair[int, str], (0, ""))
    # I decided to only match exact types because of the following issue:
    # If I have a converter for Pair[int, T] and an other for Pair[T, str],
    # which one should I choose for Pair [int, str] ...
    # maybe the first/last one declared, why not, but the feature is not
    # prioritary



from dataclasses import dataclass
from typing import Sequence, cast

from pytest import fixture, raises

from apischema import ValidationError, validator
from apischema.validation.mock import NonTrivialDependency, ValidatorMock
from apischema.validation.validator import (Validator, get_validators, validate)


@dataclass
class Data:
    a: int
    other: str = ""

    @validator
    def a_gt_10(self):
        if self.a <= 10:
            yield "error"

    @validator
    def a_lt_100(self):
        if self.a >= 100:
            raise ValueError("error2")

    @validator
    def non_trivial(self):
        non_trivial(self)


def non_trivial(data: Data):
    return data.other


@fixture
def validators() -> Sequence[Validator]:
    return get_validators(Data)


def test_get_validators(validators):
    assert validators == (Data.a_gt_10, Data.a_lt_100, Data.non_trivial)


def test_validator_descriptor():
    # Class field is descriptor
    val = cast(Validator, Data.a_gt_10)
    assert val.dependencies == {"a"}
    assert val.can_be_called({"a"})
    assert not val.can_be_called(set())
    # Can be called from class and instance
    with raises(ValueError):
        assert Data(200).a_lt_100()
    with raises(ValueError):
        assert Data.a_lt_100(Data(200))


def test_validate(validators):
    validate(Data(42), validators)
    with raises(ValidationError) as err:
        validate(Data(0), validators)
    assert err.value == ValidationError(["error"])
    with raises(ValidationError) as err:
        validate(Data(200), validators)
    assert err.value == ValidationError(["[ValueError]error2"])


def test_non_trivial(validators):
    with raises(NonTrivialDependency) as err:
        validate(ValidatorMock(Data, {"a": 42}, {}), validators)
    assert err.value.attr == "other"
    assert err.value.validator == Data.non_trivial



import re
from typing import Optional, Pattern, Union

from apischema.types import DictWithUnion, Metadata
from apischema.utils import PREFIX

PROPERTIES_METADATA = f"{PREFIX}properties"


def properties(pattern: Optional[Union[str, Pattern]] = None) -> Metadata:
    if pattern is not None:
        pattern = re.compile(pattern)
    return DictWithUnion({PROPERTIES_METADATA: pattern})




from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, Generator, Iterable, List, Set

from pytest import raises

from apischema import (ValidationError, build_input_schema,
                       field_input_converter, field_output_converter, from_data,
                       to_data, with_fields_set)


def branches_dict_from_list(branches: List[Tree]
                            ) -> Generator[Any, Any, Dict[str, Tree]]:
    nodes: Set[str] = set()
    for i, tree in enumerate(branches):
        nodes2 = set(tree.all_nodes)
        if nodes & nodes2:
            yield i, "duplicate nodes"
        nodes |= nodes2
    return {b.node: b for b in branches}


# Roughly equivalent to branch_dict_from_list
def branches_dict_from_list2(branches: List[Tree]) -> Dict[str, Tree]:
    nodes: Set[str] = set()
    errors: Dict[str, ValidationError] = {}
    for i, tree in enumerate(branches):
        nodes2 = set(tree.all_nodes)
        if nodes & nodes2:
            errors[str(i)] = ValidationError(["duplicate nodes"])
        nodes |= nodes2
    if errors:
        raise ValidationError(children=errors)
    return {b.node: b for b in branches}


def branches_dict_to_list(branches: Dict[str, Tree]) -> List[Tree]:
    return list(branches.values())


@with_fields_set
@dataclass
class Tree:
    node: str
    branches: Dict[str, Tree] = field(default_factory=dict, metadata=(
            field_input_converter(branches_dict_from_list) |
            field_output_converter(branches_dict_to_list)
    ))

    @property
    def all_nodes(self) -> Iterable[str]:
        yield self.node
        for tree in self.branches.values():
            yield tree.node
            yield from tree.all_nodes


def test_tree():
    data = {
        "node":     "root",
        "branches": [
            {"node": "leaf1"},
            {
                "node":     "node1",
                "branches": [
                    {"node": "leaf2"},
                    {"node": "leaf3"}
                ]
            }
        ]
    }
    tree = from_data(Tree, data)
    assert tree == Tree("root", {
        "leaf1": Tree("leaf1"),
        "node1": Tree("node1", {
            "leaf2": Tree("leaf2"),
            "leaf3": Tree("leaf3"),
        }),
    })
    assert to_data(tree) == data

    def ref_factory(cls: type) -> str:
        return f"#/definitions/{cls.__name__}"

    with raises(TypeError):
        build_input_schema(Tree)
    assert to_data(build_input_schema(Tree, ref_factory=ref_factory)) == {
        "additionalProperties": False,
        "properties":           {
            "node":     {"type": "string"},
            "branches": {
                "items":   {"$ref": "#/definitions/Tree"},
                "type":    "array"
            }
        },
        "required":             ["node"],
        "type":                 "object"
    }


def test_bad_tree():
    data = {
        "node":     "root",
        "branches": [
            {"node": "leaf1"},
            {
                "node":     "node1",
                "branches": [
                    {"node": "leaf2"},
                    {
                        "node":     "node2",
                        "branches": [{"node": "leaf2"}]
                    }
                ]
            }
        ]
    }
    with raises(ValidationError) as err:
        from_data(Tree, data)
    assert err.value == ValidationError(children={
        "branches": ValidationError(children={
            "1": ValidationError(children={
                "branches": ValidationError(children={
                    "1": ValidationError(["duplicate nodes"])
                })
            })
        })
    })



from dataclasses import Field as BaseField, is_dataclass
from enum import Enum
from typing import (Any, Callable, Dict, Iterable, List, Mapping, Optional,
                    Sequence, Tuple, Type, TypeVar)

from apischema.conversion import Converter, InputVisitorMixin
from apischema.data.coercion import STR_NONE_VALUES, coerce
from apischema.data.common_errors import bad_literal, wrong_type
from apischema.dataclasses import Field, FieldKind, get_input_fields
from apischema.ignore import Ignored
from apischema.schema import Constraint, Schema
from apischema.schema.constraints import (ArrayConstraint, ObjectConstraint,
                                          get_constraint)
from apischema.types import ITERABLE_TYPES, MAPPING_TYPES, NoneType
from apischema.utils import distinct
from apischema.validation.errors import (ValidationError,
                                         exception,
                                         merge)
from apischema.validation.mock import ValidatorMock
from apischema.validation.validator import Validator, get_validators, validate
from apischema.visitor import Visitor


def check_type(data: Any, expected: Type):
    if type(data) is not expected:
        raise ValidationError([wrong_type(type(data), expected)])


def validate_with_errors(data: Any, constraint: Optional[Constraint],
                         errors: Mapping[str, ValidationError]):
    if constraint is not None:
        try:
            constraint.validate(data)
        except ValidationError as err:
            raise ValidationError(err.messages, errors)
    if errors:
        raise ValidationError(children=errors)


To = TypeVar("To")
From = TypeVar("From")


def apply_converter(value: From, converter: Callable[[From], To]) -> To:
    try:
        result = converter(value)
    except ValidationError:
        raise
    except Exception as err:
        raise ValidationError([exception(err)])
    validate(result)
    return result


DataWithConstraint = Tuple[Any, Optional[Constraint]]


class FromData(InputVisitorMixin[DataWithConstraint, Any],
               Visitor[DataWithConstraint, Any]):
    def __init__(self, additional_properties: bool):
        Visitor.__init__(self)  # type: ignore
        self.additional_properties = additional_properties

    def primitive(self, cls: Type, data2: DataWithConstraint):
        data, constraint = data2
        check_type(data, cls)
        if constraint is not None and data is not None:
            constraint.validate(data)
        return data

    def _union(self, alternatives: Iterable[Type], data2: DataWithConstraint
               ) -> Tuple[Any, int]:
        error: Optional[ValidationError] = None
        for i, cls in enumerate(alternatives):
            try:
                return self.visit(cls, data2), i
            except Ignored:
                pass
            except ValidationError as err:
                error = merge(error, err)
        raise error if error is not None else TypeError("Ignored Union")

    def union(self, alternatives: Sequence[Type], data2: DataWithConstraint):
        # Optional optimization
        if data2[0] is None and alternatives[1] is NoneType:
            return None
        return self._union(alternatives, data2)[0]

    def iterable(self, cls: Type[Iterable], value_type: Type,
                 data2: DataWithConstraint):
        data, constraint = data2
        check_type(data, list)
        elts: List[value_type] = []  # type: ignore
        errors: Dict[str, ValidationError] = {}
        elt_constraint = None
        if constraint is not None:
            assert isinstance(constraint, ArrayConstraint)
            elt_constraint = constraint.items
        assert elt_constraint is None or isinstance(elt_constraint, Constraint)
        for i, elt in enumerate(data):
            try:
                elts.append(self.visit(value_type, (elt, elt_constraint)))
            except ValidationError as err:
                errors[str(i)] = err
        validate_with_errors(data, constraint, errors)
        return ITERABLE_TYPES[cls](elts)

    def mapping(self, cls: Type[Mapping], key_type: Type,  # type: ignore
                value_type: Type, data2: DataWithConstraint):
        data, constraint = data2
        check_type(data, dict)
        mapping: Dict[key_type, value_type] = {}  # type: ignore
        errors: Dict[str, ValidationError] = {}
        prop_constraint = None
        if constraint is not None:
            assert isinstance(constraint, ObjectConstraint)
            prop_constraint = constraint.additional_properties
        for key, value in data.items():
            assert isinstance(key, str)
            try:
                new_key = self.visit(key_type, (key, None))
                mapping[new_key] = self.visit(value_type,
                                              (value, prop_constraint))
            except ValidationError as err:
                errors[key] = err
        validate_with_errors(data, constraint, errors)
        return MAPPING_TYPES[cls](mapping)

    def typed_dict(self, cls: Type, keys: Mapping[str, Type],
                   total: bool, data2: DataWithConstraint):
        data, constraint = data2
        check_type(data, dict)
        constraint = constraint or get_constraint(cls)
        mapping: Dict[str, Any] = {}
        keys_in_dict = 0
        errors: Dict[str, ValidationError] = {}
        for key, value in data.items():
            assert isinstance(key, str)
            if key in keys:
                keys_in_dict += 1
                try:
                    mapping[key] = self.visit(keys[key], (value, None))
                except ValidationError as err:
                    errors[key] = err
            else:
                mapping[key] = value
        if total and keys_in_dict != len(keys):
            for key in keys:
                if key not in data:
                    errors[key] = ValidationError(["missing key"])
        validate_with_errors(data, constraint, errors)
        return mapping

    def tuple(self, types: Sequence[Type], data2: DataWithConstraint):
        data, constraint = data2
        check_type(data, list)
        ArrayConstraint(min_items=len(types),
                        max_items=len(types)).validate(data)
        elts: List[Any] = []
        errors: Dict[str, ValidationError] = {}
        elt_constraints = None
        if constraint is not None:
            assert isinstance(constraint, ArrayConstraint)
            elt_constraints = constraint.items
        for i, (cls, elt) in enumerate(zip(types, data)):
            elt_constraint: Optional[Constraint]
            if elt_constraints is None:
                elt_constraint = None
            elif isinstance(elt_constraints, Sequence):
                elt_constraint = elt_constraints[i]
            else:
                elt_constraint = elt_constraints
            try:
                elts.append(self.visit(cls, (elt, elt_constraint)))
            except ValidationError as err:
                errors[str(i)] = err
        validate_with_errors(data, constraint, errors)
        return tuple(elts)

    def literal(self, values: Sequence[Any], data2: DataWithConstraint):
        # Literal can contain Enum values and thus has to call visit too
        assert data2[1] is None
        for cls in distinct(map(type, values)):
            try:
                value = self.visit(cls, data2)
            except ValidationError:
                continue
            if value in values:
                return value
        raise ValidationError([bad_literal(data2[0], values)])

    def _custom(self, cls: Type, custom: Dict[Type, Converter],
                data2: DataWithConstraint):
        data, constraint = data2
        constraint = constraint or get_constraint(cls)
        types = list(custom)
        value, index = self._union(types, (data, constraint))
        return apply_converter(value, custom[types[index]])

    def dataclass(self, cls: Type, data2: DataWithConstraint):
        assert is_dataclass(cls)
        data, constraint = data2
        constraint = constraint or get_constraint(cls)
        check_type(data, dict)
        values: Dict[str, Any] = {}
        default: Dict[str, BaseField] = {}
        aliases: List[str] = []
        field_errors: Dict[str, ValidationError] = {}

        def set_field(field: Field, value: Any, alias: str):
            try:
                result = self.visit(field.input_type, (value, field.constraint))
                if field.input_converter:
                    result = apply_converter(result, field.input_converter)
            except ValidationError as err:
                field_errors[alias] = err
            else:
                values[field.name] = result

        fields, pattern_fields, additional_field = get_input_fields(cls)
        for field in fields:
            alias = field.alias
            if alias in data:
                aliases.append(alias)
                set_field(field, data[alias], alias)
            elif field.default:
                default[field.name] = field.base_field
            else:
                field_errors[alias] = ValidationError(["missing field"])
        if len(data) != len(aliases):
            remain = set(data).difference(aliases)
        else:
            remain = set()
        for pattern, field in pattern_fields:
            matched = {key: data[key] for key in remain if pattern.match(key)}
            remain -= matched.keys()
            set_field(field, matched, f"/{pattern.pattern}/")
        if additional_field is not None:
            additional = {key: data[key] for key in remain}
            set_field(additional_field, additional, "<additionalProperties>")
        elif remain and not self.additional_properties:
            field_errors.update((field.alias, ValidationError(["field not allowed"]))
                                for field in sorted(remain))
        error: Optional[ValidationError] = None
        if field_errors:
            error = ValidationError(children=field_errors)
        if constraint is not None:
            assert isinstance(constraint, ObjectConstraint)
            assert constraint.additional_properties is None
            try:
                constraint.validate(data)
            except ValidationError as err:
                error = merge(error, err)
        init: Dict[str, Any] = {}
        validators: Sequence[Validator]
        if hasattr(cls, "__post_init__") or error:
            for field in fields:
                if field.kind == FieldKind.INIT and field.name in values:
                    init[field.name] = values.pop(field.name)
            partial: List[Validator] = []
            whole: List[Validator] = []
            for val in get_validators(cls):
                if val.can_be_called(values.keys()):
                    partial.append(val)
                else:
                    whole.append(val)
            try:
                validate(ValidatorMock(cls, values, default), partial, **init)
            except ValidationError as err:
                error = merge(error, err)
            if error:
                raise error
            validators = whole
        else:
            validators = get_validators(cls)
        try:
            res = cls(**values, **init)
        except ValidationError:
            raise
        except Exception as err:
            raise ValidationError([exception(err)])
        validate(res, validators, **init)
        return res

    def enum(self, cls: Type[Enum], data2: DataWithConstraint):
        data, constraint = data2
        assert constraint is None
        try:
            return cls(data)
        except ValueError as err:
            raise ValidationError([str(err)])

    def new_type(self, cls: Type, super_type: Type, data2: DataWithConstraint):
        data, constraint = data2
        data2_ = data, constraint or get_constraint(cls)
        return self.visit(super_type, data2_)

    def any(self, data2: DataWithConstraint):
        return data2[0]

    def annotated(self, cls: Type, annotations: Sequence[Any],
                  data2: DataWithConstraint):
        if Ignored in annotations:
            raise Ignored()
        data, constraint = data2
        if constraint is None:
            for annotation in annotations:
                if isinstance(annotation, Schema):
                    constraint = annotation.constraint
                    break
        return self.visit(cls, (data, constraint))


T = TypeVar("T")


class FromDataWithCoercion(FromData):
    def __init__(self, additional_properties: bool):
        super().__init__(additional_properties)

    def primitive(self, cls, data2: DataWithConstraint):
        data, constraint = data2
        if cls is NoneType or data is None:
            if data not in STR_NONE_VALUES:
                check_type(data, cls)
            return None
        data = coerce(cls, data)
        if constraint is not None and data is not None:
            constraint.validate(data)
        return data

    def enum(self, cls: Type[Enum], data2: DataWithConstraint):
        data, constraint = data2
        data = self.primitive(type(data), (data, None))
        return super().enum(cls, (data, constraint))


def from_data(cls: Type[T], data: Any, *, additional_properties: bool = False,
              coerce: bool = False) -> T:
    visitor = FromData if not coerce else FromDataWithCoercion
    return visitor(additional_properties).visit(cls, (data, None))



from _pytest.python_api import raises

from apischema.data import items_to_data


def test_build_data():
    lines = {
        "key1.0": "v0",
        "key1.2": "v2",
        "key2":   42,
    }
    assert items_to_data(lines.items()) == {"key1": ["v0", None, "v2"],
                                            "key2": 42}
    with raises(ValueError):
        items_to_data({"": ...}.items())
    with raises(ValueError):
        items_to_data({**lines, "key1.key3": ...}.items())
    with raises(ValueError):
        items_to_data({**lines, "0": ...}.items())



import re
from dataclasses import dataclass
from typing import Any, Dict, Iterator, Optional, Pattern, Sequence, TypeVar, Union

from apischema.types import Number
from apischema.utils import PREFIX, to_hashable
from apischema.validation.errors import ValidationError

CONSTRAINT_METADATA = f"{PREFIX}constraint"

T = TypeVar("T")


@dataclass
class Constraint:
    read_only: Optional[bool] = None
    write_only: Optional[bool] = None

    def __post_init__(self):
        if self.read_only is not None and self.write_only is not None:
            raise ValueError("Schema cannot be both read/write only")

    def _validate(self, data) -> Iterator[str]:
        raise NotImplementedError()

    def validate(self, data: T) -> T:
        errors = list(self._validate(data))
        if errors:
            raise ValidationError(errors)
        return data


_constraints: Dict[Any, Constraint] = {}

get_constraint = _constraints.get

Cls = TypeVar("Cls", bound=type)


@dataclass
class NumberConstraint(Constraint):
    minimum: Optional[Number] = None
    maximum: Optional[Number] = None
    exclusive_minimum: Optional[Number] = None
    exclusive_maximum: Optional[Number] = None
    multiple_of: Optional[Number] = None

    def _validate(self, data: Any) -> Iterator[str]:
        assert isinstance(data, (int, float))
        if self.minimum is not None and data < self.minimum:
            yield f"{data} < {self.minimum} (minimum)"
        if self.maximum is not None and data > self.maximum:
            yield f"{data} > {self.maximum} (maximum)"
        if self.exclusive_minimum is not None:
            if data <= self.exclusive_minimum:
                yield f"{data} <= {self.minimum} (exclusiveMinimum)"
        if self.exclusive_maximum is not None:
            if data >= self.exclusive_maximum:
                yield f"{data} >= {self.maximum} (exclusiveMaximum)"
        if self.multiple_of is not None and (data % self.multiple_of) != 0:
            yield (f"{data} if not a multiple of "
                   f"{self.multiple_of} (multipleOf)")


@dataclass
class StringConstraint(Constraint):
    min_length: Optional[int] = None
    max_length: Optional[int] = None
    pattern: Optional[Union[str, Pattern]] = None

    def __post_init__(self):
        super().__post_init__()
        if self.pattern is not None:
            self.pattern = re.compile(self.pattern)

    def _validate(self, data: Any) -> Iterator[str]:
        assert isinstance(data, str)
        if self.min_length is not None and len(data) < self.min_length:
            yield f"'{data}'.length < {self.min_length} (minLength)"
        if self.max_length is not None and len(data) > self.max_length:
            yield f"'{data}'.length > {self.max_length} (maxLength)"
        if self.pattern is not None and not re.fullmatch(self.pattern, data):
            yield f"'{data}' does not match '{self.pattern}' pattern"


@dataclass
class ArrayConstraint(Constraint):
    min_items: Optional[int] = None
    max_items: Optional[int] = None
    items: Optional[Union[Constraint, Sequence[Constraint]]] = None
    unique_items: Optional[bool] = None

    def _validate(self, data: Any) -> Iterator[str]:
        assert isinstance(data, list)
        if self.min_items is not None and len(data) < self.min_items:
            yield (f"not enough items, {len(data)} is lower than "
                   f"{self.min_items} (minItems)")
        if self.max_items is not None and len(data) > self.max_items:
            yield (f"too much items, {len(data)} is greater than "
                   f"{self.max_items} (maxItems)")
        if self.unique_items and len(set(map(to_hashable, data))) != len(data):
            yield f"duplicates items in {data} (uniqueItems)"


@dataclass
class ObjectConstraint(Constraint):
    min_properties: Optional[int] = None
    max_properties: Optional[int] = None
    additional_properties: Optional[Constraint] = None

    def _validate(self, data: Any) -> Iterator[str]:
        assert isinstance(data, dict)
        if self.min_properties is not None and len(data) < self.min_properties:
            yield (f"not enough properties, {len(data)} is lower than "
                   f"{self.min_properties} (minProperties)")
        if self.max_properties is not None and len(data) > self.max_properties:
            yield (f"too much properties, {len(data)} is greater than "
                   f"{self.max_properties} (maxProperties)")



import re
import sys
from datetime import date, datetime, time
from ipaddress import (IPv4Address, IPv4Interface, IPv4Network,
                       IPv6Address, IPv6Interface, IPv6Network)
from typing import Pattern
from uuid import UUID

from apischema.conversion import inout_model, input_converter, output_converter
from apischema.schema import schema

# ==================== uuid ====================

inout_model(str)(UUID)
schema(format="uuid")(UUID)

# ==================== datetime ====================

if sys.version_info >= (3, 7):
    input_converter(date.fromisoformat, str, date)
    output_converter(date.isoformat, date, str)
    schema(format="date")(date)

    input_converter(datetime.fromisoformat, str, datetime)
    output_converter(datetime.isoformat, datetime, str)
    schema(format="date-time")(datetime)

    input_converter(time.fromisoformat, str, time)
    input_converter(time.isoformat, time, str)
    schema(format="time")(time)

# ==================== ipaddress ====================

inout_model(str)(IPv4Address)
schema(format="ipv4")(IPv4Address)

inout_model(str)(IPv4Interface)
schema(format="ipv4")(IPv4Interface)

inout_model(str)(IPv4Network)
schema(format="ipv4")(IPv4Network)

inout_model(str)(IPv6Address)
schema(format="ipv6")(IPv6Address)

inout_model(str)(IPv6Interface)
schema(format="ipv6")(IPv6Interface)

inout_model(str)(IPv6Network)
schema(format="ipv6")(IPv6Network)

# ==================== pattern ====================
input_converter(re.compile, str, Pattern)
output_converter(lambda p: p.pattern, Pattern, str)



from dataclasses import dataclass, field
from typing import List, Tuple

from pytest import raises

from apischema import Discard, ValidationError, from_data, get_fields, validator


@dataclass
class Result:
    bounds: Tuple[int, int] = field()
    values: List[int]

    @validator
    def bounds_are_sorted(self):
        min_bound, max_bound = self.bounds
        if min_bound > max_bound:
            fields = get_fields(self)
            yield fields.bounds, "bounds are not sorted"
            raise Discard(fields.bounds)

    @validator(bounds)
    def bounds_are_sorted2(self):  # equivalent to bounds_are_sorted
        min_bound, max_bound = self.bounds
        if min_bound > max_bound:
            yield "bounds are not sorted"

    @validator
    def values_dont_exceed_bounds(self):
        min_bound, max_bound = self.bounds
        for index, value in enumerate(self.values):
            if not min_bound <= value <= max_bound:
                yield index, f"value exceeds bounds"


def test_result():
    data = {
        "bounds": [0, 10],
        "values": [1, 3, 4],
    }
    assert from_data(Result, data) == Result((0, 10), [1, 3, 4])


def test_bad_bounds():
    data = {
        "bounds": [10, 0],
        "values": [1, 3, 4],
    }
    with raises(ValidationError) as err:
        from_data(Result, data)
    assert err.value == ValidationError(children={
        "bounds": ValidationError(["bounds are not sorted"])
    })


def test_bad_values():
    data = {
        "bounds": [0, 10],
        "values": [1, 3, -4, 4, 42],
    }
    with raises(ValidationError) as err:
        from_data(Result, data)
    assert err.value == ValidationError(children={
        "2": ValidationError(["value exceeds bounds"]),
        "4": ValidationError(["value exceeds bounds"]),
    })



from base64 import b64decode, b64encode
from typing import NewType

from pytest import raises

from apischema import ValidationError, from_data, input_converter
from apischema.conversion import raw_input_converter

input_converter(b64decode, str, bytes)
CheckedBinary = NewType("CheckedBinary", bytes)


def compute_checksum(bytes_: bytes) -> int:
    return sum(bytes_) % (2 ** 32)


@raw_input_converter
def check_binary(binary: bytes, checksum: int) -> CheckedBinary:
    if checksum != compute_checksum(binary):
        raise ValidationError(["invalid checksum"])
    return CheckedBinary(binary)


def test_check_binary():
    binary = b'data!'
    data = {
        "binary":   b64encode(binary).decode(),
        "checksum": compute_checksum(binary),
    }
    assert from_data(CheckedBinary, data) == binary
    with raises(ValidationError) as err:
        from_data(CheckedBinary, {"binary": b64encode(binary).decode(), "checksum": 0})
    assert err.value == ValidationError(["invalid checksum"])



from enum import Enum
from typing import (Any, Dict, Iterable, Mapping, Sequence, Tuple,
                    Type,
                    TypeVar)

from apischema.conversion import (Converter, OutputVisitorMixin)
from apischema.data.common_errors import bad_literal, wrong_type
from apischema.dataclasses import Field, get_output_fields_raw
from apischema.fields import get_fields_set
from apischema.types import UNTYPED_COLLECTIONS
from apischema.visitor import Visitor

UNTYPED_COLLECTIONS_IDS = set(map(id, UNTYPED_COLLECTIONS))


def check_type(obj: Any, expected: Type):
    if not isinstance(obj, expected):
        raise ValueError(wrong_type(type(obj), expected))


def check_mapping(mapping: Mapping):
    if not all(isinstance(k, str) for k in mapping):
        raise ValueError("mapping keys must be strings")
    return mapping


class ToData(OutputVisitorMixin[Any, Any], Visitor[Any, Any]):
    def __init__(self, conversions: Mapping[Type, Type], exclude_unset: bool):
        Visitor.__init__(self)
        OutputVisitorMixin.__init__(self, conversions)
        self.exclude_unset = exclude_unset

    def primitive(self, cls: Type, obj):
        check_type(obj, cls)
        return obj

    def union(self, alternatives: Sequence[Type], obj):
        for cls in alternatives:
            try:
                return self.visit(cls, obj)
            except:  # noqa
                pass
        raise ValueError(f"No union alternatives {list(alternatives)}"
                         f" matches '{type(obj)}'")

    def iterable(self, cls: Type[Iterable], value_type: Type, obj):
        check_type(obj, Iterable)
        return [self.visit(value_type, elt) for elt in obj]

    def mapping(self, cls: Type[Mapping], key_type: Type,
                value_type: Type, obj):
        check_type(obj, Mapping)
        return check_mapping({
            self.visit(key_type, key): self.visit(value_type, value)
            for key, value in obj.items()
        })

    def typed_dict(self, cls: Type, keys: Mapping[str, Type],
                   total: bool, obj):
        check_type(obj, Mapping)
        if total and any(key not in obj for key in keys):
            raise ValueError(f"Typed '{cls}' is not total: {obj}")
        return {
            key: self.visit(keys[key], value) if key in keys else value
            for key, value in obj.items()
        }

    def tuple(self, types: Sequence[Type], obj):
        check_type(obj, tuple)
        if len(obj) != len(types):
            raise ValueError(f"Expected tuple length '{len(types)}',"
                             f" got '{len(obj)}'")
        return [self.visit(cls, elt) for cls, elt in zip(types, obj)]

    def literal(self, values: Sequence[Any], obj):
        if obj not in values:
            raise ValueError(bad_literal(obj, values))
        return obj

    def _custom(self, cls: Type, custom: Tuple[Type, Converter], obj):
        conv_type, converter = custom
        return self.visit(conv_type, converter(obj))

    def dataclass(self, cls: Type, obj):
        result: Dict[str, Any] = {}

        def field_value(field: Field):
            value = getattr(obj, field.name)
            if field.output_converter is not None:
                value = field.output_converter(value)
            return value

        fields, properties_fields = get_output_fields_raw(cls)
        if self.exclude_unset:
            fields_set = get_fields_set(obj)
            fields = [f for f in fields if f.name in fields_set]
        for field in fields:
            value = self.visit(field.output_type, field_value(field))
            result[field.alias] = value
        for field in properties_fields:
            value = self.visit(field.output_type, field_value(field))
            result.update(value)
        return result

    def enum(self, cls: Type[Enum], obj):
        check_type(obj, cls)
        return obj.value

    def any(self, obj):
        if isinstance(obj, Mapping):
            return check_mapping({
                self.visit(type(k), k): self.visit(type(v), v)
                for k, v in obj.items()
            })
        if isinstance(obj, str):
            return self.visit(type(obj), obj)
        if isinstance(obj, Iterable):
            return [self.visit(type(v), v) for v in obj]
        return self.visit(type(obj), obj)

    def visit(self, cls: Type, obj):
        # Use `id` to avoid useless costly generic types hashing
        if id(cls) in UNTYPED_COLLECTIONS_IDS:
            return self.visit(UNTYPED_COLLECTIONS[cls], obj)  # type: ignore
        return super().visit(cls, obj)


T = TypeVar("T")


def to_data(obj: T, cls: Type[T] = None,
            conversions: Mapping[Type, Type] = None,
            *, exclude_unset: bool = True) -> Any:
    if cls is None:
        cls = type(obj)
    if conversions is None:
        conversions = {}
    return ToData(conversions, exclude_unset).visit(cls, obj)



from dataclasses import dataclass, field
from enum import Enum
from typing import (AbstractSet, Any, List, Mapping, Optional,
                    Sequence, Set, Union)
from uuid import UUID, uuid4

from pytest import mark, raises

from apischema import properties, schema
from apischema.conversion import inout_model
from apischema.data import from_data
from apischema.data.to_data import to_data
from apischema.fields import with_fields_set
from apischema.typing import Literal
from apischema.validation.errors import ValidationError
from apischema.validation.validator import validator

uuid = str(uuid4())


def bijection(cls, data, expected):
    obj = from_data(cls, data)
    assert obj == expected
    assert to_data(obj, cls) == data


def error(data, cls):
    with raises(ValidationError):
        from_data(cls, data)


@dataclass(unsafe_hash=True)
class SimpleDataclass:
    a: int


class SimpleEnum(Enum):
    a = "a"


@with_fields_set
@dataclass
class Dataclass:
    nested: SimpleDataclass
    opt: Optional[int] = field(default=None, metadata=schema(min=100))


@dataclass
class PartialValidator:
    a: int = 0
    b: int = 0
    c: int = 0

    @validator
    def validate(self):
        if self.a == self.b:
            yield "error"


@mark.parametrize("data", ["", 0])
def test_any(data):
    bijection(Any, data, data)


@mark.parametrize("data, expected", [
    (None, None),
    ({"a": 0}, SimpleDataclass(0)),
])
def test_optional(data, expected):
    bijection(Optional[SimpleDataclass], data, expected)


def test_optional_error():
    error(0, Optional[str])


@mark.parametrize("data, expected", [
    ("", ""),
    ({"a": 0}, SimpleDataclass(0))
])
def test_union(data, expected):
    bijection(Union[str, SimpleDataclass], data, expected)


@mark.parametrize("data", [0, None])
def test_union_error(data):
    error(data, Union[str, SimpleDataclass])


def test_union_value_error():
    with raises(ValueError):
        to_data(True, Union[str, List[str]])


@mark.parametrize("cls, data", [
    (int, 0),
    (str, ""),
    (bool, True),
    (float, 0.0)
])
def test_primitive(cls, data):
    bijection(cls, data, data)


@mark.parametrize("data", ["", None])
def test_primitive_error(data):
    error(data, int)


# noinspection PyTypeChecker
@mark.parametrize("cls, expected", [
    (List, [0, SimpleDataclass(0)]),
    (Set, {0, SimpleDataclass(0)}),
    (Sequence, (0, SimpleDataclass(0))),
    (AbstractSet, frozenset([0, SimpleDataclass(0)]))
])
def test_iterable(cls, expected):
    data = [0, {"a": 0}]
    bijection(cls[Union[int, SimpleDataclass]], data, expected)


@mark.parametrize("data", [{}, ["", 0]])
def test_iterable_error(data):
    error(data, List[str])


@mark.parametrize("key_cls, data, expected", [
    (str, {"int": 0, "SC": {"a": 0}}, {"int": 0, "SC": SimpleDataclass(0)}),
    (UUID, {uuid: 0}, {UUID(uuid): 0}),
    (UUID, {uuid: 0}, {UUID(uuid): 0}),
])
def test_mapping(key_cls, data, expected):
    bijection(Mapping[key_cls, Union[int, SimpleDataclass]], data, expected)


@mark.parametrize("data", [
    [],
    {"key": ""},
])
def test_mapping_error(data):
    error(data, Mapping[str, int])


@mark.parametrize("expected", [UUID(uuid), UUID(uuid)])
def test_model(expected):
    bijection(UUID, uuid, expected)


@mark.parametrize("data", [0, "fake"])
def test_model_error(data):
    error(data, UUID)


def test_enum():
    bijection(SimpleEnum, "a", SimpleEnum.a)


def test_enum_errors():
    error("b", SimpleEnum)


@mark.parametrize("data", [0, "ok"])
def test_literal(data):
    bijection(Literal[0, "ok"], data, data)


def test_literal_error():
    error(1, Literal[0, "ok"])
    with raises(ValueError):
        to_data(1, Literal[0, "ok"])


@mark.parametrize("data, expected", [
    ({"nested": {"a": 0}}, Dataclass(SimpleDataclass(0), None)),
    ({"nested": {"a": 0},
      "opt":    None}, Dataclass(SimpleDataclass(0), None)),
    ({"nested": {"a": 0}, "opt": 100}, Dataclass(SimpleDataclass(0), 100)),
])
def test_dataclass(data, expected):
    bijection(Dataclass, data, expected)


def test_dataclass_partial_validator():
    with raises(ValidationError):
        from_data(PartialValidator, {})


@mark.parametrize("data", [{}, {"nested": {}, "opt": 1}])
def test_dataclass_error(data):
    error(data, Dataclass)


def test_with_class_context():
    @inout_model(int)
    @schema(min=100)
    class BigInt(int):
        pass

    bijection(BigInt, 100, 100)


def test_properties():
    @dataclass
    class Test:
        startswith_a: Mapping[str, Any] = field(metadata=properties("^a.*$"))
        others: Mapping[str, Any] = field(metadata=properties())

    assert from_data(Test, {"plop": 0, "allo": 1}) == Test({"allo": 1},
                                                           {"plop": 0})



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Mapping, Optional, Pattern, Sequence, Type, Union

from apischema.alias import alias
from apischema.fields import with_fields_set
from apischema.types import NoneType, Number
from apischema.utils import NO_DEFAULT, to_camel_case


class JSONType(Enum):
    NULL = "null"
    BOOLEAN = "boolean"
    STRING = "string"
    INTEGER = "integer"
    NUMBER = "number"
    ARRAY = "array"
    OBJECT = "object"

    @staticmethod
    def from_type(cls: Type) -> "JSONType":
        return {
            NoneType: JSONType.NULL,
            bool:     JSONType.BOOLEAN,
            str:      JSONType.STRING,
            int:      JSONType.INTEGER,
            float:    JSONType.NUMBER,
            list:     JSONType.ARRAY,
            dict:     JSONType.OBJECT,
        }[cls]


@alias(to_camel_case)
@with_fields_set
@dataclass
class JSONSchema:
    additional_properties: Optional[Union[bool, "JSONSchema"]] = None
    all_of: Optional[Sequence["JSONSchema"]] = None
    any_of: Optional[Sequence["JSONSchema"]] = None
    const: Any = NO_DEFAULT
    default: Any = NO_DEFAULT
    description: Optional[str] = None
    enum: Optional[Sequence[Any]] = None
    exclusive_maximum: Optional[Number] = None
    exclusive_minimum: Optional[Number] = None
    examples: Optional[Sequence[Any]] = None
    format: Optional[str] = None
    items: Optional[Union["JSONSchema", Sequence["JSONSchema"]]] = None
    maximum: Optional[Number] = None
    minimum: Optional[Number] = None
    max_items: Optional[int] = None
    min_items: Optional[int] = None
    max_length: Optional[int] = None
    min_length: Optional[int] = None
    max_properties: Optional[int] = None
    min_properties: Optional[int] = None
    multiple_of: Optional[Number] = None
    one_of: Optional[Sequence["JSONSchema"]] = None
    pattern: Optional[Union[str, Pattern]] = None
    pattern_properties: Optional[Mapping[Union[str, Pattern], "JSONSchema"]] = None
    properties: Optional[Mapping[str, "JSONSchema"]] = None
    read_only: Optional[bool] = None
    ref: Optional[str] = field(default=None, metadata=alias("$ref"))
    required: Optional[Sequence[str]] = None
    title: Optional[str] = None
    type: Optional[Union[JSONType, Sequence[JSONType]]] = None
    unique_items: Optional[bool] = None
    write_only: Optional[bool] = None
